#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>


#define PAGE_SIZE	4096
#define MARGIN		10	/* if things get really nasty, we need this */

struct thread {
    long *pages;	/* array of pages we've obtained */
    int num_pages;	/* number of pages in "pages" */
    pthread_t thread;
};


/* a tad racy on stdio, but who cares */

static void *reader(void *arg)
{
    struct thread *self = arg;
    long page[PAGE_SIZE/sizeof(long)];

    while (1) {
	ssize_t got;

	got = read(0,page,PAGE_SIZE);
	if (!got) break;
	if (got < 0) {
	    perror("read");
	    return (void *) 1;
	}
	if (got != PAGE_SIZE) {
	    fprintf(stderr,"short read %ld\n",(long) got);
	    return (void *) 1;
	}
	self->pages[self->num_pages++] = page[0];
    }
    return NULL;
}


static void usage(const char *name)
{
    fprintf(stderr,"usage: %s NUM_THREADS <FILE\n",name);
    exit(1);
}


int main(int argc,char **argv)
{
    long num_threads;
    char *end;
    struct stat st;
    struct thread *threads;
    int i,j,sum;

    if (argc != 2) usage(*argv);
    num_threads = strtol(argv[1],&end,0);
    if (*end || num_threads < 0) usage(*argv);
    if (fstat(0,&st) < 0) {
	perror("fstat");
	return 1;
    }

    /* allocate memory */
    threads = malloc(num_threads*sizeof(struct thread));
    if (!threads) {
	perror("malloc");
	return 1;
    }
    for (i = 0; i != num_threads; i++) {
	threads[i].pages = malloc(st.st_size/PAGE_SIZE*sizeof(long)*MARGIN);
	if (!threads[i].pages) {
	    perror("malloc");
	    return 1;
	}
	threads[i].num_pages = 0;
    }

    /* create and run all threads */
    fprintf(stderr,"launching %ld thread(s)\n",num_threads);
    for (i = 0; i != num_threads; i++)
	if (pthread_create(&threads[i].thread,NULL,reader,threads+i)) {
	    perror("pthread_create");
	    return 1;
	}

    /* wait for them to finish */
    for (i = 0; i != num_threads; i++) {
	void *res;

	if (pthread_join(threads[i].thread,&res)) {
	    perror("pthread_join");
	    return 1;
	}
	if (res) {
	    fprintf(stderr,"thread %d failed\n",i);
	    return 1;
	}
    }

    /* count what we got */
    fprintf(stderr,"checking");
    sum = 0;
    for (i = 0; i != num_threads; i++) {
	fprintf(stderr,"%c%d",i ? '+' : ' ',threads[i].num_pages);
	sum += threads[i].num_pages;
    }
    fprintf(stderr,"=%d/%d pages\n",sum,(int) (st.st_size/PAGE_SIZE));

    /* check that each page is present */
    for (i = 0; i != st.st_size/PAGE_SIZE; i++) {
	int found = 0;

	for (j = 0; j != num_threads; j++)
	    if (threads[j].num_pages && *threads[j].pages == i) {
		found++;
		threads[j].pages++;
		threads[j].num_pages--;
	    }
	if (!found) {
	    fprintf(stderr,"page %d not found\n",i);
	    return 1;
	}
	if (found > 1) {
	    fprintf(stderr,"page %d found %d times\n",i,found);
	    return 1;
	}
    }

    fprintf(stderr,"OK\n");
    return 0;
}
